/*==============================================================================


       eg_17 - Database handling

               Copyright  2001 Pankhurst Algorithmics - All Rights Reserved
               visit www.ebmDevMag.com for permissions/restrictions on use

==============================================================================*/
#include <string.h>

#include "malloc.h"
#include "gui.h"
#include "OpenDatabases.h"
#include "FieldMgr.h"

extern "C"
{
#include "ereader_hostio.h"
}
#include "eg_17_form.h"
/*==============================================================================

       database & file manager objects

==============================================================================*/
COpenDB m_db;
CFMDatabase m_fMgr;
/*==============================================================================

       database calls (wrappers)

==============================================================================*/
DBErr DatabaseCreate(const ebo_name_t *fileName,FieldDef*fieldDef,long maxSize)
{
  // create brand-new db
	CDBHeaderHdl dbh;
	DBErr error=m_db.Create(0,(ebo_name_t*)fileName,OS_availaddr,maxSize,&dbh);
  if ( DB_OUT_OF_MEMORY == error ) // no room?
  {
    GUI_NeedFileSpace(fileName); // ask OS for more space
    error=m_db.Create(0,(ebo_name_t*)fileName,OS_availaddr,maxSize,&dbh); // and try again
    if ( DB_OUT_OF_MEMORY == error ) // still no room?
      return error;
  }                                                                             
  return m_fMgr.InitFieldDefinitions(&m_db,fieldDef,0,0); // set up manager
}
//------------------------------------------------------------------------------
DBErr DatabaseOpen(const ebo_name_t *fileName,FieldDef *fieldDef,long maxSize,int *existed)
{
  // open database 'fileName' - if nonexistent, create it also
  DBErr error=DB_OK;
	if (m_db.Open( 0, (ebo_name_t *) fileName, OS_availaddr, maxSize ))
  {
    *existed=0;
    error=DatabaseCreate(fileName,fieldDef,maxSize);
    if ( DB_OK != error )
      return error;
	}
  else
    *existed=1; // opened successfully - already there
  error=m_fMgr.Open(&m_db);
	if (!error) // success?
    OS_availaddr+=maxSize; // adjust memory so we don't reuse addresses assigned to db
  return error;
}
//------------------------------------------------------------------------------
int DatabaseNoOfRec(void)
{
  // for consistency
  return m_db.GetNumRecords();
}
//------------------------------------------------------------------------------
DBErr DatabaseClose(void)
{
  // obviously, close db
  DBErr error=m_fMgr.Close(); // close file manager
  if ( DB_OK == error ) // ok closing?
    error=m_db.Close(); // close and set error flag on db
  return error;
}
/*==============================================================================

       record calls - makes working with records more straightforward

==============================================================================*/
DBErr RecordDelete(int recNo)
{
  // delete specific record
  return m_db.DeleteRecord(recNo);
}
//------------------------------------------------------------------------------
DBErr RecordAdd(const ebo_name_t *fileName,char**field)
{
  // add array of data to new record - 'field' is an array of field entries, one
  // per field, matching EXACTLY, as no error checking done!
  CFMRecord record(&m_fMgr);
  DBErr error=record.New();
  if ( DB_OK!=error ) // unable to set aside new record?
    return error;
  // get fields needed from database - note that field array passed in had
  // better be the same number, so production code should use assertions here
  int totalField=m_fMgr.GetNumFields();
  for (int fieldNo=0;fieldNo<totalField;fieldNo++)
  {
    // add field, acquiring space as needed
    if (record.PutField(fieldNo,(U8*)field[fieldNo])) // failed?
    {
      // need more space - ask OS for it, and error if not enough provided
      GUI_NeedFileSpace( fileName );
      if ( 0 != (error=record.PutField(fieldNo,(U8*)field[fieldNo])) ) // try again
        return error;
    }
  }
  if ( DB_OUT_OF_MEMORY==record.Add() ) // no memory to add rec?
  {
    // again, ask OS for space, and try again
    GUI_NeedFileSpace( fileName );
    error=record.Add();
    if ( DB_OUT_OF_MEMORY==error )
      return error;
  }
  return record.Close();
}
//------------------------------------------------------------------------------
DBErr RecordCopyToArray(int recNo,char**array)
{
  // read 'recNo' into array passed - note that the strings in 'array'
  // are allocated by 'new', and so 1) need to be 'delete[]'ed after, 2) if there
  // is any data already allocated by the array's pointers, the pointers to that
  // data will be lost, so only pass a 'clean' array in!
  char *ptr=0;
  CFMRecord record(&m_fMgr);
  DBErr error=record.Open(recNo);
  if ( error )
    return error;
  int totalField=m_fMgr.GetNumFields();
  for (int fieldNo=0;fieldNo<totalField;fieldNo++)
  {
    record.GetField(fieldNo,(U8**)&ptr);
    array[fieldNo]=new char[strlen(ptr)+1]; // copy each field into a char array
    strcpy(array[fieldNo],ptr);
  }
  return error;
}
//------------------------------------------------------------------------------
DBErr RecordCopyFromArray(const ebo_name_t *fileName,int recNo,char**array)
{
  // copy strings in 'array' into record 'recNo' - no memory management is
  // done, so array will have to be cleaned up by caller routine
  CFMRecord record(&m_fMgr);
  DBErr error=record.Open(recNo);
  if ( error )
    return error;
  int totalField=m_fMgr.GetNumFields();
  for (int fieldNo=0;fieldNo<totalField;fieldNo++)
  {
    if ( DB_OUT_OF_MEMORY==record.PutField(fieldNo,(U8 *)array[fieldNo])) // need more room?
    {
      GUI_NeedFileSpace(fileName); // ask OS & try again
      error=record.PutField(fieldNo,(U8 *)array[fieldNo]);
      if ( DB_OUT_OF_MEMORY==error )
        return error;
    }
  }
  return error;
}
/*==============================================================================

       main window routines - all routines below work on the database via the
       above routines, and do not access the database or db variables directly

==============================================================================*/
// file definition - what is seen in the ebm file manager
const ebo_name_t s_myDatabase = { "",          // security - leave empty
                                  "Developer", // *your* development name
                                  "MyDB",      // file name
                                  "db" };      // file extension
//------------------------------------------------------------------------------
// field definitions - we state field type, then maximum size (not including the
// null terminator) and field name, with FLD_EOF to mark end of definitions
FieldDef m_fieldDef[] = {
    { FLD_TEXT, 64, "Animal"   },
    { FLD_TEXT, 60, "Location" },
    { FLD_TEXT, 32, "Type"     },
    { FLD_EOF,   0,          0 } }; // FLD_EOF always ends definition!
//------------------------------------------------------------------------------
// so our code always remains in sync with db, we get array size as indicator of
// # of fields in file - note we subtract one from the value since we don't count
// the last entry (FLD_EOF)
const int c_noOfField=sizeof(m_fieldDef)/sizeof(m_fieldDef[0]) - 1;
//------------------------------------------------------------------------------
// sample data written in when AddDatabaseData() is called
char *m_sampleData[][3] = {
        { "Eagle",       "Air",   "Bird"    },
        { "Aardvark",    "Land",  "Mammal"  },
        { "Coral Snake", "Land",  "Reptile" },
        { "Dolphin",     "Ocean", "Mammal"  }
                          };
//------------------------------------------------------------------------------
bool CWindow1::AddDatabaseData(void)
{
  // append the above data records to the database
  int totalRec=sizeof(m_sampleData)/sizeof(m_sampleData[0]);
  for (int recNo=0;recNo<totalRec;recNo++)
  {
    if ( RecordAdd(&s_myDatabase,m_sampleData[recNo]) )
      return false;
  }
  DisplayRecord(0);
  return true;
}
//------------------------------------------------------------------------------
void CWindow1::DisplayRecord(int index)
{
  // this displays a selected record onscreen
  if ( index<0 || index>=DatabaseNoOfRec() )
    return; // nothing to show...
  char *array[c_noOfField];
  static char caption[99]="";
  caption[0]='\0';
  if (DB_OK==RecordCopyToArray(index,array))
  {
    // adjust label
    sprintf(caption,"Rec #%d",index);
    // add entries to display
    FIELD_0->SetText(array[0]);
    FIELD_1->SetText(array[1]);
    FIELD_2->SetText(array[2]);
    for (int i=0;i<c_noOfField;i++)
      delete[] array[i];
    m_currRec=index;
  }
  LABEL_REC->SetLabel(caption);
}
//------------------------------------------------------------------------------
bool CWindow1::SaveRecord(int index)
{
  // write the data on the display into record 'index'
  char *array[c_noOfField];
  array[0]=FIELD_0->GetText(); // get data from each entry
  array[1]=FIELD_1->GetText();
  array[2]=FIELD_2->GetText();
  DBErr error=RecordCopyFromArray(&s_myDatabase,index,array);
  for (int i=0;i<c_noOfField;i++)
    free(array[i]);
  return error;
}
//------------------------------------------------------------------------------
bool CWindow1::AddNewRecord(void)
{
  // write the data on the display into a new record
  char *array[c_noOfField];
  array[0]=FIELD_0->GetText(); // get data from each entry
  array[1]=FIELD_1->GetText();
  array[2]=FIELD_2->GetText();
  DBErr error=RecordAdd(&s_myDatabase,array); // add to end of db
  DisplayRecord(DatabaseNoOfRec()-1); // display last entry (this record)
  for (int i=0;i<c_noOfField;i++)
    free(array[i]);
  return error;
}
//------------------------------------------------------------------------------
S32 CWindow1::MsgHandler( MSG_TYPE type, CViewable *from, S32 data )
{
  // process button/radio button messages
  if ( MSG_BUTTON_SELECT==type ) // show dialog?
  {
    switch (from->GetID())
    {
      case CWindow1_BUTTON_OPEN:
      {
        int existed=0;
        if ( DB_OK != DatabaseOpen(&s_myDatabase,m_fieldDef,131072,&existed) )
          GUI_Alert( ALERT_WARNING, "Unable to open database!" );
        else
        {
          if (existed) // already existing?
            GUI_Alert( ALERT_WARNING, "Database reopened" );
          else
            GUI_Alert( ALERT_WARNING, "New database created and opened" );
          DisplayRecord(0);
        }
      }
        break;
      case CWindow1_BUTTON_CREATE:
        if ( ! AddDatabaseData() )
          GUI_Alert( ALERT_WARNING, "Unable to add to database!" );
        else
          GUI_Alert( ALERT_WARNING, "Data added to database" );
        break;
      case CWindow1_BUTTON_CLOSE:
        if ( DB_OK != DatabaseClose() )
          GUI_Alert( ALERT_WARNING, "Unable to close database!" );
        else
          GUI_Alert( ALERT_WARNING, "Database closed" );
        break;
      case CWindow1_BUTTON_PREV_REC:
        if (m_currRec>=1)
          DisplayRecord(m_currRec-1);
        break;
      case CWindow1_BUTTON_NEXT_REC:
        if (m_currRec+1<DatabaseNoOfRec())
          DisplayRecord(m_currRec+1);
        break;
      case CWindow1_BUTTON_DEL_REC:
        if ( DB_OK != RecordDelete(m_currRec) )
          GUI_Alert( ALERT_WARNING, "Unable to delete record!" );
        else
        {
          GUI_Alert( ALERT_WARNING, "Record deleted" );
          DisplayRecord(0);
        }
        break;
      case CWindow1_BUTTON_SAVE_REC:
        if ( DB_OK != SaveRecord(m_currRec) )
          GUI_Alert( ALERT_WARNING, "Unable to save record!" );
        else
          GUI_Alert( ALERT_WARNING, "Record saved" );
        break;
      case CWindow1_BUTTON_ADD_NEW_REC:
        if ( DB_OK != AddNewRecord() )
          GUI_Alert( ALERT_WARNING, "Unable to save new record!" );
        else
          GUI_Alert( ALERT_WARNING, "New record saved" );
        break;
    }
  }
    return 1;
  return CWindow::MsgHandler(type, from, data);
}
/*==============================================================================


==============================================================================*/
static CWindow1 *f_mainWindow;
//------------------------------------------------------------------------------
S32 GUI_main( MSG_TYPE type, CViewable *object, S32 data )
{
  switch (type)
  {
    case MSG_APP_START:
      if (OS_is_present && hostIO_am_I_the_current_task())
        HOSTIO_INLINE_BREAKPOINT();
      f_mainWindow = new CWindow1();
      return 1;
    case MSG_APP_STOP:
      delete f_mainWindow;
      return 1;
    default:
      break;
  }
  return 0;
};
//------------------------------------------------------------------------------



